import re
import typing as T
import snowfakery.data_gen_exceptions as exc
from dateutil.rrule import rrule, rruleset
from dateutil import rrule as rrule_mod
from datetime import datetime, date, time, timezone
from snowfakery import PluginResultIterator
from snowfakery.plugins import SnowfakeryPlugin, memorable
from snowfakery.template_funcs import parse_datetimespec, parse_date


# Note

#   As soon as the base rule and the exclusion differ by even a second,
#   the exclusion doesn't exclude properly. (I think microseconds are okay.)
#
#   E.g. if I exclude very Monday in December, but the "Monday" is Monday at 3:30
#        and the days are scheduled for 3:31, then the exclusion won't apply.

FREQ_STRS = {
    frequency: getattr(rrule_mod, frequency)
    for frequency in [
        "YEARLY",
        "MONTHLY",
        "WEEKLY",
        "DAILY",
        "HOURLY",
        "MINUTELY",
        "SECONDLY",
    ]
}


WEEKDAYS = {
    day: getattr(rrule_mod, day) for day in ["MO", "TU", "WE", "TH", "FR", "SA", "SU"]
}


def is_datetime(dt: str) -> bool:
    """Does a string look like a `datetime`?"""
    return bool(set(dt).intersection(" TZ+:"))


OptionalDateTimeLike = T.Union[str, date, datetime, None]
SeqOfIntsLike = T.Union[str, int, T.Sequence[int], None]
CalendarException = T.Union[
    T.Sequence[T.Union["CalendarRule", OptionalDateTimeLike]],
    "CalendarRule",
    OptionalDateTimeLike,
]


class CalendarRule(PluginResultIterator):
    """Generates calendar events"""

    iterator: T.Any  # Actually T.Union[T.Iterator[date],T.Iterator[ datetime]]

    def __init__(
        self,
        freq: str,
        start_date: OptionalDateTimeLike = None,
        interval: int = 1,
        count: T.Optional[int] = None,
        until: OptionalDateTimeLike = None,
        bysetpos: SeqOfIntsLike = None,  # undocumented feature, for now
        bymonth: SeqOfIntsLike = None,
        bymonthday: SeqOfIntsLike = None,
        byyearday: SeqOfIntsLike = None,
        byeaster: SeqOfIntsLike = None,  # undocumented feature for now
        byweekno: SeqOfIntsLike = None,  # undocumented feature for now
        byweekday: T.Optional[str] = None,
        byhour: SeqOfIntsLike = None,
        byminute: SeqOfIntsLike = None,
        bysecond: SeqOfIntsLike = None,
        cache: bool = False,  # undocumented feature for now
        exclude: CalendarException = None,
        include: CalendarException = None,
        use_undocumented_features: bool = False,
    ):
        # Certain features can only be used if you opt-in to use_undocumented_features
        self._check_undocumented_features(
            use_undocumented_features, bysetpos, byeaster, cache, byweekno
        )

        # Normalize and save the start_date
        self.start_date, precision = self._normalize_start_date(start_date)

        self._set_output_datetype_date_or_datetime(precision)

        wkst = rrule_mod.SU  # Standardize on Sunday as start of week globally
        # Could make this parameterizable if there is demand for it.
        bysetpos = process_list_of_ints(bysetpos)
        bymonth = process_list_of_ints(bymonth)
        bymonthday = process_list_of_ints(bymonthday)
        byyearday = process_list_of_ints(byyearday)
        byeaster = process_list_of_ints(byeaster)
        byhour = process_list_of_ints(byhour)
        byminute = process_list_of_ints(byminute)
        bysecond = process_list_of_ints(bysecond)
        byweekno = process_list_of_ints(bysecond)

        until = self._normalize_until(until)

        freq = self._normalize_frequency(freq)

        if byweekday:
            byweekday_normalized = self._normalize_weekday(byweekday)
        else:
            byweekday_normalized = None

        self.ruleset = rruleset(cache)
        self.rrule = rrule(
            freq=freq,
            dtstart=self.start_date,
            interval=interval,
            wkst=wkst,
            count=count,
            until=until,
            bysetpos=bysetpos,  # undocumented feature for now
            bymonth=bymonth,  # undocumented feature for now
            bymonthday=bymonthday,
            byyearday=byyearday,
            byeaster=byeaster,  # undocumented feature for now
            byweekno=byweekno,
            byweekday=byweekday_normalized,
            byhour=byhour,
            byminute=byminute,
            bysecond=bysecond,
            cache=cache,
        )
        self.ruleset.rrule(self.rrule)

        self.compound = include or exclude
        if exclude:
            self._process_special_cases(exclude, "exclude")

        if include:
            self._process_special_cases(include, "include")

        self.iterator = iter(self)

    def _check_undocumented_features(
        self,
        use_undocumented_features: bool,
        bysetpos: SeqOfIntsLike,
        byeaster: SeqOfIntsLike,
        cache: bool,
        byweekno: SeqOfIntsLike,
    ) -> None:
        """Require opt-in for undocumented features"""
        if (not use_undocumented_features) and any(
            [bysetpos, byeaster, cache, byweekno]
        ):
            raise exc.DataGenValueError(
                "That feature is undocumented and unsupported. "
                "Use the `use_undocumented_features: True` argument to use it regardless."
            )

    def _normalize_start_date(
        self, start_date: OptionalDateTimeLike
    ) -> T.Tuple[datetime, type]:
        """Normalize the start-date to datetime

        But return a type that keeps track of its original type"""

        if isinstance(start_date, str):
            precision = datetime if is_datetime(start_date) else date
            start_date = parse_datetimespec(start_date)
            assert start_date.tzinfo
        elif isinstance(start_date, datetime):
            precision = datetime
        elif isinstance(start_date, date):
            start_date = datetime.combine(
                start_date, time(hour=0, minute=0, second=0, tzinfo=timezone.utc)
            )
            precision = date
        elif not start_date:
            start_date = datetime.now()
            precision = datetime
        else:  # pragma: no cover
            raise TypeError(
                f"`start_date` ({start_date}) is of unknown type {type(start_date)}"
            )
        if not start_date.tzinfo:
            start_date = start_date.replace(tzinfo=timezone.utc)
        assert precision in (date, datetime)

        return start_date, precision

    def _normalize_until(self, until: OptionalDateTimeLike) -> T.Optional[datetime]:
        """Normalize until value to datetime or None"""
        if not until:
            return None

        if isinstance(until, str):
            if is_datetime(until):
                until = parse_datetimespec(until)
            else:
                until = datetime.combine(parse_date(until), self.start_date.time())

        elif isinstance(until, date):
            until = datetime.combine(until, self.start_date.time(), tzinfo=timezone.utc)

        else:
            raise exc.DataGenTypeError(
                f"`until` parameter ({until}) is of unexpected type {until}"
            )

        return until.replace(tzinfo=timezone.utc)

    def _set_output_datetype_date_or_datetime(self, precision: type) -> None:
        """Depending on the precision requested, generate the right kinds of records"""
        assert precision in (date, datetime)
        if precision is date:
            self.next = self._next_date  # type: ignore
        elif precision is datetime:
            self.next = self._next_datetime  # type: ignore
        else:  # pragma: no cover
            assert False, "Should be unreachable"

    def _process_special_cases(
        self,
        case: CalendarException,
        action: T.Literal["include", "exclude"],
    ) -> None:
        """Process inclusions and exclusions"""
        if action == "exclude":
            # https://dateutil.readthedocs.io/en/stable/rrule.html#dateutil.rrule.rruleset.exrule
            add_rule = self.ruleset.exrule
            # https://dateutil.readthedocs.io/en/stable/rrule.html#dateutil.rrule.rruleset.exdate
            add_date = self.ruleset.exdate
        elif action == "include":
            # https://dateutil.readthedocs.io/en/stable/rrule.html#dateutil.rrule.rruleset.rrule
            add_rule = self.ruleset.rrule
            # https://dateutil.readthedocs.io/en/stable/rrule.html#dateutil.rrule.rruleset.rdate
            add_date = self.ruleset.rdate
        else:  # pragma: no cover   -   Should be unreachable
            assert action in ("include", "exclude"), "Bad action!"
            raise NotImplementedError("Bad action!")

        if isinstance(case, (list, tuple)):
            for case in case:
                self._process_special_cases(case, action)
        elif isinstance(case, CalendarRule):
            add_rule(T.cast(T.Any, case.ruleset))

        elif isinstance(case, datetime):
            add_date(case)
        elif isinstance(case, date):
            d: date = case
            self._process_special_cases(
                datetime.combine(d, self.start_date.time(), tzinfo=timezone.utc), action
            )
        elif isinstance(case, str):
            d2: date = parse_date(case)
            dt: datetime = datetime.combine(
                d2, self.start_date.time(), tzinfo=timezone.utc
            )
            self._process_special_cases(dt, action)
        else:  # pragma: no cover
            raise TypeError(f"Cannot {action} {case}, ({type(case)})")

    def _normalize_frequency(self, freq: str):
        """Normalize frequency string to dateutil type"""
        try:
            freq_str = freq.upper()
            freq = FREQ_STRS[freq_str]
        except KeyError:
            raise exc.DataGenError(
                f"Cannot accept frequency string '{freq}'. Valid strings are {tuple(FREQ_STRS.keys())}"
            )

        except Exception as e:
            raise exc.DataGenError(
                f"Cannot accept frequency string '{freq}' because {repr(e)}"
            )
        if freq in (rrule_mod.HOURLY, rrule_mod.MINUTELY, rrule_mod.SECONDLY):

            if self.next != self._next_datetime:
                raise exc.DataGenError(
                    "If frequency is a time value, `start_date` should be a datetime"
                )

        return freq

    def _normalize_weekday(
        self, byweekday: T.Optional[str]
    ) -> T.List[rrule_mod.weekday]:
        """Normalize weekday string to dateutil types"""
        assert byweekday

        if isinstance(byweekday, str):
            weekdays = byweekday.split(",")
        else:
            raise exc.DataGenTypeError(
                f"Wrong type for byweekday {byweekday}, {type(byweekday)}"
            )

        days = [self._parse_weekday(day) for day in weekdays]

        return days

    def _parse_weekday(self, day):
        """Parse the weekday syntax like MO(+3) and WE(-2) and create dateutil datatypes"""
        try:
            if "(" in day:
                day, offset = self._split_weekday(day)
            else:
                offset = None
            dayconst = WEEKDAYS[day.upper().strip()]

        except (TypeError, KeyError):
            raise exc.DataGenTypeError(
                f"Do not know how to interpret {day} as a day of the week."
            )

        if offset:
            dayconst = dayconst(offset)
        return dayconst

    def _split_weekday(self, day) -> T.Tuple[str, int]:
        """Parse the weekday syntax like MO(+3) and WE(-2)"""
        parts = re.match(r"\s*(?P<weekday>\w+)\((?P<offset>.+)\)\s*", day)
        if not parts:
            raise exc.DataGenSyntaxError(f"Cannot understand {day}")
        weekday = parts["weekday"]
        offset = parts["offset"]
        try:
            offset = int(offset)
        except ValueError:
            raise exc.DataGenTypeError(f"Cannot interpret {offset} as a number")
        return weekday, offset

    def __iter__(self) -> T.Union[T.Iterator[datetime], T.Iterator[date]]:
        return iter(self.ruleset)

    def next(self) -> None:  # pragma: no cover
        """This method is never called.

        It is replaced at runtime by _next_datetime or _next_date"""
        raise NotImplementedError("next is not implemented")

    def _next_datetime(self) -> datetime:
        return next(self.iterator)

    def _next_date(self) -> date:
        val: datetime = next(self.iterator)
        return val.date()


class Schedule(SnowfakeryPlugin):
    class Functions:
        @memorable
        def Event(
            self,
            freq: str,
            start_date=None,
            interval=1,
            # wkst=None,
            count=None,
            until=None,
            bysetpos=None,  # undocumented feature for now
            bymonth=None,  # undocumented feature for now
            bymonthday=None,
            byyearday=None,
            byeaster=None,
            byweekno=None,
            byweekday=None,
            byhour=None,
            byminute=None,
            bysecond=None,
            cache=False,
            exclude=None,
            include=None,
        ):
            return CalendarRule(
                freq=freq,
                start_date=start_date,
                interval=interval,
                # wkst=wkst,
                count=count,
                until=until,
                bysetpos=bysetpos,  # No requirement identified yet
                bymonth=bymonth,  # No requirement id
                bymonthday=bymonthday,
                byyearday=byyearday,
                byeaster=byeaster,
                byweekno=byweekno,
                byweekday=byweekday,
                byhour=byhour,
                byminute=byminute,
                bysecond=bysecond,
                cache=cache,
                exclude=exclude,
                include=include,
            )


def process_list_of_ints(val: SeqOfIntsLike) -> T.Optional[T.List[int]]:
    if val is None:
        return None
    elif isinstance(val, int):
        return [val]
    elif isinstance(val, str):
        return [int(v) for v in val.split(",")]
    elif isinstance(val, (list, tuple)):
        return [int(v) for v in val]
    else:
        raise exc.DataGenTypeError(
            f"Expected a number or list of numbers, not {val} ({type(val)})"
        )



import sqlite3
import typing as T
import warnings
from collections import defaultdict
from copy import deepcopy
from random import randint

from snowfakery import data_gen_exceptions as exc
from snowfakery.object_rows import (
    LazyLoadedObjectReference,
    NicknameSlot,
    ObjectReference,
    ObjectRow,
)
from snowfakery.plugins import PluginResultIterator
from snowfakery.utils.pickle import RestrictedPickler
from snowfakery.utils.randomized_range import UpdatableRandomRange


class RowHistory:
    """Remember tables that might be random_reference'd in a database."""

    already_warned = False

    def __init__(
        self,
        table_counters: T.Mapping,
        tables_to_keep_history_for: T.Iterable[str],
        tablename_for_nickname: T.Mapping[str, str],
    ):
        self.conn = sqlite3.connect("")
        self.table_counters = dict(table_counters)
        self.nickname_counters = defaultdict(int)
        self.reset_locals()
        # the pattern is A -> A means A is a table
        #                B -> A means B is a nickname and A is a table
        #
        self.nickname_to_tablename = {
            nick: table
            for nick, table in tablename_for_nickname.items()
            if table != nick
        }
        for table in tables_to_keep_history_for:
            _make_history_table(self.conn, table)
        self.pickler = RestrictedPickler(_DISPATCH_TABLE, _SAFE_CLASSES)

    def reset_locals(self):
        """Reset the minimum count that counts as "local" """
        self.local_counters = deepcopy(self.table_counters)

    def save_row(self, tablename: str, nickname: T.Optional[str], row: dict):
        """Save a row to temporary storage"""
        row_id = row["id"]

        # keep track of highest ID
        self.table_counters[tablename] = row_id

        if nickname:
            nickname_id = self._get_nickname_id(tablename, nickname)
            self.table_counters[nickname] = nickname_id
        else:
            nickname_id = None

        # note that this dumps a full object tree
        # that will cause some duplication of data but doing a big
        # "join" across multiple tables would have other costs (even if done lazily).
        # For now this seems best and simplest.
        # The data de-dupling algorithm would be slightly complex and slow.
        data = self.pickler.dumps(row)
        self.conn.execute(
            f'INSERT INTO "{tablename}" VALUES (?, ?, ?, ?)',
            (row_id, nickname, nickname_id, data),
        )

    def random_row_reference(self, name: str, scope: str, randomizer_func: callable):
        """Find a random row and load it"""
        if scope not in ("prior-and-current-iterations", "current-iteration"):
            raise exc.DataGenError(
                f"Scope must be 'prior-and-current-iterations' or 'current-iteration' not {scope}",
                None,
                None,
            )

        # Next Step: implement "unique" with a Linear Congruent Generator

        if name in self.nickname_to_tablename:
            nickname = name
            tablename = self.nickname_to_tablename[nickname]
            max_id = self.nickname_counters[nickname]
        else:
            nickname = None
            tablename = name
            max_id = self.table_counters.get(tablename)

        if not max_id:
            raise exc.DataGenError(
                f"There is no table or nickname `{nickname or tablename}` at this point in the recipe."
            )

        if scope == "prior-and-current-iterations":
            if not self.already_warned:
                warnings.warn("Global scope is an experimental feature.")
                self.already_warned = True
            min_id = 1
        elif nickname:
            min_id = self.local_counters.get(nickname, 0) + 1
        else:
            min_id = self.local_counters.get(tablename, 0) + 1
        # if no records can be found in this iteration
        # just look through the whole table.
        #
        # This happens usually when you are referring to just_once
        if max_id < min_id:
            min_id = 1

        if nickname:
            # find a random nickname'd row by its nickname_id
            nickname_id = randomizer_func(min_id, max_id)
            row_id = self.find_row_id_for_nickname_id(tablename, nickname, nickname_id)
        else:
            # find a random row
            row_id = randomizer_func(min_id, max_id)

        return LazyLoadedObjectReference(tablename, row_id, tablename)

    def load_row(self, tablename: str, row_id: int):
        """Load a row from the DB by row_id/object_id"""
        qr = self.conn.execute(
            f'SELECT DATA FROM "{tablename}" WHERE id=?',
            (row_id,),
        )
        first_row = next(qr, None)
        assert first_row, f"Something went wrong: we cannot find {tablename}: {row_id}"

        return self.pickler.loads(first_row[0])

    def find_row_id_for_nickname_id(
        self, tablename: str, nickname: str, nickname_id: int
    ):
        #     """Find a nicknamed row by its nickname_id"""
        qr = self.conn.execute(
            f'SELECT id FROM "{tablename}" WHERE nickname=? AND nickname_id=?',
            (nickname, nickname_id),
        )
        first_row = next(qr, None)
        assert (
            first_row
        ), f"Something went wrong: we cannot find {tablename}: {nickname} : {nickname_id}"

        return first_row[0]

    def _get_nickname_id(self, tablename: str, nickname: str):
        """Get a unique auto-incrementing nickname identifier for a new row"""
        self.nickname_counters[nickname] += 1
        return self.nickname_counters[nickname]


def _make_history_table(conn, tablename):
    """Make a history table"""

    c = conn.cursor()
    c.execute(
        f'CREATE TABLE "{tablename}" (id INTEGER NOT NULL UNIQUE, nickname VARCHAR, nickname_id INTEGER, data VARCHAR NOT NULL)'
    )
    # helps with sparsely scattered nicknames. Of debatable value. Can speed up benchmarks
    # but hard to see it in real recipes.
    c.execute(
        f'CREATE UNIQUE INDEX "{tablename}_nickname_id" ON "{tablename}" (nickname, nickname_id);'
    )


_DISPATCH_TABLE = {
    NicknameSlot: lambda n: (
        ObjectReference,
        (n._tablename, n.allocated_id),
    ),
    ObjectRow: lambda v: (
        ObjectRow,
        (v._tablename, v._values),
    ),
}

_SAFE_CLASSES = {
    ("snowfakery.object_rows", "ObjectRow"),
    ("snowfakery.object_rows", "ObjectReference"),
    ("snowfakery.object_rows", "LazyLoadedObjectReference"),
    ("decimal", "Decimal"),
    ("datetime", "date"),
    ("datetime", "datetime"),
    ("datetime", "timedelta"),
    ("datetime", "timezone"),
}


class RandomReferenceContext(PluginResultIterator):
    # initialize the object's state.
    rng = None

    def __init__(
        self,
        row_history: RowHistory,
        to: str,
        scope: str = "current-iteration",
        unique: bool = False,
    ):
        self.row_history = row_history
        self.to = to
        self.scope = scope
        self.unique = unique
        if unique:
            self.random_func = self.unique_random
        else:
            self.random_func = randint

    def next(self) -> T.Union[ObjectReference, ObjectRow]:
        try:
            return self.row_history.random_row_reference(
                self.to, self.scope, self.random_func
            )
        except StopIteration as e:
            if self.random_func == self.unique_random:
                raise exc.DataGenError(
                    f"Cannot find an unused `{self.to}`` to link to"
                ) from e
            else:  # pragma: no cover
                raise e

    def unique_random(self, a, b):
        """Goal: use an Uniquifying RNG until all of its values have been
        used up, then make a new one, with higher values.

        e.g. random_range(1,5) then random_range(5, 10)

        The parent might call it like:
        unique_random(1,2) -> random_range(1,3) -> 2
        unique_random(1,4) -> random_range(1,3) -> 1
        unique_random(1,6) -> random_range(3,7) -> 5  # reset
        unique_random(1,8) -> random_range(3,7) -> 3
        unique_random(1,10) -> random_range(3,7) -> 4
        unique_random(1,12) -> random_range(3,7) -> 6
        unique_random(1,14) -> random_range(7,14) -> 13 # reset
        ...
        """
        b += 1  # randomizer_func uses top-inclusive semantics,
        # random_range uses top-exclusive semantics
        if self.rng is None:
            self.rng = UpdatableRandomRange(a, b)
        else:
            self.rng.set_new_range(a, b)
        return next(self.rng)



from abc import abstractmethod, ABC
from .data_generator_runtime import evaluate_function, RuntimeContext, Interpreter
from .object_rows import ObjectRow, ObjectReference
from contextlib import contextmanager
from typing import NamedTuple, Union, Dict, Sequence, Optional, cast
from .utils.template_utils import look_for_number
import itertools
import jinja2

from .data_gen_exceptions import (
    DataGenError,
    DataGenNameError,
    DataGenSyntaxError,
    DataGenValueError,
    fix_exception,
)
from .plugins import Scalar, PluginResult, PluginResultIterator

# objects that represent the hierarchy of a data generator.
# roughly similar to the YAML structure but with domain-specific objects
Definition = Union["ObjectTemplate", "SimpleValue", "StructuredValue"]
FieldValue = Union[None, Scalar, ObjectRow, tuple, PluginResult, ObjectReference]


class FieldDefinition(ABC):
    """Base class for things that render fields

    Abstract base class for everything that can fulfill the role of X in

    - object: abc
      fields:
         fieldname: X
    """

    @abstractmethod
    def render(self, context: RuntimeContext) -> FieldValue:
        pass

    @contextmanager
    def exception_handling(self, message: str, *args, **kwargs):
        try:
            yield
        except Exception as e:
            raise fix_exception(message, self, e, *args, **kwargs) from e


class VariableDefinition:
    """Defines a mutable variable. Like:

    - var: Foo
      value: Bar
    """

    # TODO: Add an example

    tablename = None

    def __init__(
        self,
        filename: str,
        line_num: int,
        varname: str,
        expression: Definition,
    ):
        self.varname = varname
        self.expression = expression
        self.filename = filename
        self.line_num = line_num

    def evaluate(self, context: RuntimeContext) -> FieldValue:
        """Evaluate the expression"""
        return self.expression.render(context)

    def execute(
        self, interp: Interpreter, parent_context: RuntimeContext, continuing: bool
    ) -> Optional[dict]:
        with parent_context.child_context(self) as context:
            name = self.varname
            value = self.evaluate(context)
        interp.register_variable(name, value)


class ForEachVariableDefinition:
    """Represents a for_each statement."""

    varname: str
    expression: "StructuredValue"

    def __init__(
        self, filename: str, line_num: int, varname: str, expression: Definition
    ):
        self.varname = varname
        self.expression = expression
        self.filename = filename
        self.line_num = line_num

    def evaluate(self, context: RuntimeContext) -> FieldValue:
        """Disable value caching for this context and evaluate the expression"""
        context.recalculate_every_time = True
        ret = self.expression.render(context)
        if not isinstance(ret, PluginResultIterator):
            raise DataGenValueError(
                f"`for_each` value must be a DatasetIterator for `{self.varname}`",
                self.filename,
                self.line_num,
            )
        ret.repeat = False
        return ret


class LoopIterator(NamedTuple):
    name: str
    iterator: object


class ObjectTemplate:
    """A factory that generates rows.

    The runtime equivalent of

    - object: tablename
      count: count_expr   # counts can be dynamic so they are expressions
      fields: list of FieldFactories
      friends: list of other ObjectTemplates
      nickname: string
    """

    def __init__(
        self,
        tablename: str,
        filename: str,
        line_num: int,
        nickname: str = None,
        count_expr: FieldDefinition = None,  # counts can be dynamic so they are expressions
        for_each_expr: ForEachVariableDefinition = None,
        just_once: bool = False,
        fields: Sequence = (),
        friends: Sequence = (),
        update_key: str = None,
    ):
        self.tablename = tablename
        self.nickname = nickname
        self.count_expr = count_expr
        self.just_once = just_once
        self.filename = filename
        self.line_num = line_num
        self.fields = fields
        self.friends = friends
        self.for_each_expr = for_each_expr
        self.update_key = update_key

        if count_expr and for_each_expr:
            raise DataGenSyntaxError(
                f"Cannot specify both a count expression and a for-each expression at the same time in declaration for {self.tablename}.",
                self.filename,
                self.line_num,
            )

    def render(self, context: RuntimeContext) -> Optional[ObjectRow]:
        return self.generate_rows(context.output_stream, context)

    def generate_rows(
        self, output_stream, parent_context: RuntimeContext
    ) -> Optional[ObjectRow]:
        """Generate several rows"""
        rc = None
        with parent_context.child_context(self) as context:
            if self.for_each_expr:
                # it would be easy to support multiple parallel
                # for-eaches here and at one point the code did,
                # but the use-case wasn't obvious so it was removed
                # after 9bc296a7df. If we get to 2023 without
                # # finding a use-case we can delete this comment.
                iterators = [self._evaluate_for_each(context)]
                iterators.append(LoopIterator("child_index", itertools.count()))
            else:  # use a count, or a default count of 1
                iterators = [
                    LoopIterator(
                        "child_index", iter(range(self._evaluate_count(context)))
                    )
                ]
            with self.exception_handling(f"Cannot generate {self.name}"):
                master_iterator = zip(*(it.iterator for it in iterators))
                iterator_names = [it.name for it in iterators]
                for i, next_value_list in enumerate(master_iterator):
                    for name, value in zip(iterator_names, next_value_list):
                        context.interpreter.register_variable(name, value)
                    rc = self._generate_row(output_stream, context, i)

        return rc  # return last row

    @contextmanager
    def exception_handling(self, message: str):
        try:
            yield
        except DataGenError:
            raise
        except Exception as e:
            raise DataGenError(
                f"{message} : {str(e)}", self.filename, self.line_num
            ) from e

    def _evaluate_count(self, context: RuntimeContext) -> int:
        """Evaluate the count expression to an integer"""
        if not self.count_expr:
            return 1
        else:
            try:
                return int(float(cast(str, self.count_expr.render(context))))
            except (ValueError, TypeError) as e:
                raise DataGenValueError(
                    f"Cannot evaluate {self.count_expr.definition} as number",
                    self.count_expr.filename,
                    self.count_expr.line_num,
                ) from e

    @property
    def name(self) -> str:
        name = self.tablename
        if self.nickname:
            name += f" ({self.nickname})"
        return name

    @property
    def id(self):
        return id(self)

    def _evaluate_for_each(self, context: RuntimeContext) -> LoopIterator:
        """Evaluate the expression to get an iterator we can iterate over"""

        def eval_to_iterator(vardef):
            val = vardef.evaluate(context)
            try:
                iterator = iter(val)
            except TypeError:  # pragma: no cover
                raise DataGenError(
                    f"Object created by {vardef.varname} is not iterable: {val}"
                )
            return LoopIterator(vardef.varname, iterator)

        with self.exception_handling("Cannot evaluate `for_each` definition"):
            return eval_to_iterator(self.for_each_expr)

    def _generate_row(
        self, output_stream, context: RuntimeContext, index: int
    ) -> ObjectRow:
        """Generate an individual row"""
        id = context.generate_id(self.nickname)
        row = {"id": id}

        # add a column keeping track of what update_key was specified by
        # the template. This allows multiple templates to have different
        # update_keys.
        if self.update_key:
            row["_sf_update_key"] = self.update_key
        sobj = ObjectRow(self.tablename, row, index)

        context.register_object(sobj, self.nickname, self.just_once)

        self._generate_fields(context, row)

        context.remember_row(
            self.tablename,
            self.nickname,
            row,
        )

        with self.exception_handling("Cannot write row"):
            if not self.tablename.startswith("__"):
                output_stream.write_row(self.tablename, context.filter_row_values(row))

        context.interpreter.loop_over_templates_once(self.friends, True)
        return sobj

    def _generate_fields(self, context: RuntimeContext, row: Dict) -> None:
        """Generate all of the fields of a row"""
        for field in self.fields:
            with self.exception_handling("Problem rendering value"):
                value = field.generate_value(context)
                if isinstance(value, PluginResultIterator):
                    try:
                        value = value.next()
                    except StopIteration:
                        raise DataGenError(
                            "Could not generate enough values to create rows",
                            self.filename,
                            self.line_num,
                        )
                row[field.name] = value

                self._check_type(field, row[field.name], context)

    def _check_type(self, field, generated_value, context: RuntimeContext):
        """Check the type of a field value"""
        if not isinstance(generated_value, FieldValue.__args__):
            raise DataGenValueError(
                f"Field '{field.name}' generated unexpected object: {generated_value} {type(generated_value)}",
                self.filename,
                self.line_num,
            )

    def execute(
        self, interp: Interpreter, context: RuntimeContext, continuing: bool
    ) -> Optional[ObjectRow]:
        should_skip = self.just_once and continuing
        if not should_skip:
            self.generate_rows(interp.output_stream, interp.current_context)


class SimpleValue(FieldDefinition):
    """A value with no sub-structure (although it could hold a formula)

    - object: abc
      fields:
         fieldname: XXXXX
         fieldname2: ${{XXXXX}}
         fieldname3: 42
    """

    def __init__(self, definition: Scalar, filename: str, line_num: int):
        self.filename = filename
        self.line_num = line_num
        self.definition: Scalar = definition
        assert isinstance(filename, str)
        assert isinstance(line_num, int), line_num
        self._evaluator = None

    def evaluator(self, context):
        """Populate the evaluator property once."""
        if self._evaluator is None:
            if isinstance(self.definition, str):
                with self.exception_handling("Cannot parse value {}", self.definition):
                    self._evaluator = context.get_evaluator(self.definition)
            else:
                self._evaluator = False
        return self._evaluator

    def render(self, context: RuntimeContext) -> FieldValue:
        """Render the value: rendering a template if necessary."""
        old_context_identifier = context.unique_context_identifier
        context.unique_context_identifier = str(id(self))
        evaluator = self.evaluator(context)
        if evaluator:
            try:
                val = evaluator(context)
                if hasattr(val, "render"):
                    val = val.render()
            except jinja2.exceptions.UndefinedError as e:
                raise DataGenNameError(e.message, self.filename, self.line_num) from e
            except Exception as e:
                raise DataGenValueError(str(e), self.filename, self.line_num) from e
        else:
            val = self.definition
        context.unique_context_identifier = old_context_identifier
        if isinstance(val, str) and not context.interpreter.native_types:
            val = look_for_number(val)
        return val

    def __repr__(self):
        return f"<{self.__class__.__name__ , self.definition}>"


class StructuredValue(FieldDefinition):
    """A value with substructure which will call a handler function.

    - object: abc
      fields:
        fieldname:
            - reference:
                foo
        fieldname2:
            - random_number:
                min: 10
                max: 20
        fieldname3:
            - reference:
                ..."""

    def __init__(self, function_name, args, filename, line_num):
        self.function_name = function_name
        self.filename = filename
        self.line_num = line_num
        if isinstance(args, list):  # lists will represent your arguments
            self.args = args
            self.kwargs = {}
        elif isinstance(args, dict):  # dicts will represent named arguments
            self.args = []
            self.kwargs = args
        else:  # scalars will be turned into a one-argument list
            self.args = [args]
            self.kwargs = {}
        self.unique_context_identifier = str(id(self))

    def render(self, context: RuntimeContext) -> FieldValue:
        context.unique_context_identifier = self.unique_context_identifier
        if "." in self.function_name:
            objname, method, *rest = self.function_name.split(".")
            if rest:
                raise DataGenSyntaxError(
                    f"Function names should have only one '.' in them: {self.function_name}",
                    self.filename,
                    self.line_num,
                )
            obj = context.field_vars().get(objname)
            if not obj:
                raise DataGenNameError(
                    f"Cannot find definition for: {objname}",
                    self.filename,
                    self.line_num,
                )
            try:
                func = getattr(obj, method)
            except AttributeError:
                # clean up the error message a bit
                raise AttributeError(
                    f"'{objname}' plugin exposes no attribute '{method}'"
                )
            if not callable(func):
                raise DataGenNameError(
                    f"Cannot call '{method}' on '{objname}'",
                    self.filename,
                    self.line_num,
                )
            value = evaluate_function(func, self.args, self.kwargs, context)
        else:
            try:
                func = context.executable_blocks()[self.function_name]
            except KeyError:
                raise DataGenNameError(
                    f"Cannot find function named {self.function_name} to handle field value",
                    self.filename,
                    self.line_num,
                )

            with self.exception_handling(
                "Cannot evaluate function `{}`:\n {e}", [self.function_name]
            ):
                value = evaluate_function(func, self.args, self.kwargs, context)

        return value

    def __repr__(self):
        return (
            f"<StructuredValue: {self.function_name} (*{self.args}, **{self.kwargs})>"
        )


class FieldFactory:
    """Represents a single data field (name, value) to be rendered

    - object:
      fields:
        name: value   # this part
    """

    def __init__(self, name: str, definition: Definition, filename: str, line_num: int):
        self.name = name
        self.filename = filename
        self.line_num = line_num
        self.definition = definition

    def generate_value(self, context) -> FieldValue:
        try:
            return self.definition.render(context)
        except Exception as e:
            raise fix_exception(
                "Problem rendering field {}:\n {e}", self, e, [self.name]
            ) from e

    def __repr__(self):
        return f"<{self.__class__.__name__, self.name, self.definition.__class__.__name__}>"


Statement = Union[ObjectTemplate, VariableDefinition]



from pathlib import Path
from contextlib import contextmanager, ExitStack
import typing as T
import sys

import yaml

from snowfakery.data_generator import generate

from snowfakery.output_streams import (
    DebugOutputStream,
    MultiplexOutputStream,
    SqlDbOutputStream,
)
from snowfakery.generate_mapping_from_recipe import mapping_from_recipe_templates
from snowfakery.salesforce import create_cci_record_type_tables
from snowfakery.cci_mapping_files.declaration_parser import (
    SObjectRuleDeclarationFile,
    unify,
)
import snowfakery.data_gen_exceptions as exc
from snowfakery.data_generator_runtime import (
    StoppingCriteria,
)
from snowfakery.utils.files import FileLike, open_file_like

OUTPUT_FORMATS = {
    "png": "snowfakery.output_streams.ImageOutputStream",
    "svg": "snowfakery.output_streams.ImageOutputStream",
    "svgz": "snowfakery.output_streams.ImageOutputStream",
    "jpeg": "snowfakery.output_streams.ImageOutputStream",
    "jpg": "snowfakery.output_streams.ImageOutputStream",
    "ps": "snowfakery.output_streams.ImageOutputStream",
    "dot": "snowfakery.output_streams.GraphvizOutputStream",
    "json": "snowfakery.output_streams.JSONOutputStream",
    "txt": "snowfakery.output_streams.DebugOutputStream",
    "csv": "snowfakery.output_streams.CSVOutputStream",
    "sql": "snowfakery.output_streams.SqlTextOutputStream",
}

file_extensions = tuple(OUTPUT_FORMATS.keys())

COUNT_REPS = "__REPS__"


class SnowfakeryApplication:
    """Base class for all applications which embed Snowfakery as a library,
    including the Snowfakery CLI and CumulusCI"""

    stopping_criteria = None
    starting_id = 0
    rep_count = 0

    def __init__(self, stopping_criteria: StoppingCriteria = None):
        self.stopping_criteria = stopping_criteria or StoppingCriteria(COUNT_REPS, 1)

    def echo(self, message=None, file=None, nl=True, err=False, color=None):
        """Write something to a virtual stdout or stderr.

        Arguments to this function are derived from click.echo"""
        import click

        click.echo(message, file, nl, err, color)

    @property
    def stopping_tablename(self):
        """Return the name of "stopping table/object":

        The table/object whose presence determines
        whether we are done generating data.

        This is used by Snowfakery to validate that
        the provided recipe will not generate forever
        due to a misspelling the stopping tablename."""
        if self.stopping_criteria.tablename != COUNT_REPS:
            return self.stopping_criteria.tablename

    def ensure_progress_was_made(self, id_manager):
        """Check that we are actually making progress towards our goal"""
        if not self.stopping_tablename:
            return False

        last_used_id = id_manager[self.stopping_tablename]

        if last_used_id == self.starting_id:
            raise RuntimeError(
                f"{self.stopping_tablename} max ID was {self.starting_id} "
                f"before evaluating recipe and is {last_used_id} after. "
                f"At this rate we will never hit our target!"
            )

        self.starting_id = last_used_id

    def check_if_finished(self, id_manager):
        "Check whether we've finished making as many objects as we promised"
        # if nobody told us how much to make, finish after first run
        self.rep_count += 1

        target_table, count = self.stopping_criteria

        if target_table == COUNT_REPS:
            return self.rep_count >= count

        # Snowfakery processes can be restarted. We would need
        # to keep track of where we restarted to know whether
        # we are truly finished
        start = id_manager.start_ids.get(target_table, 1)
        target_id = start + count - 1

        last_used_id = id_manager[target_table]

        return last_used_id >= target_id


def stopping_criteria_from_target_number(target_number):
    "Deconstruct a tuple of 'str number' or 'number str' and make a StoppingCriteria"

    # 'number str' is the official format so the other one can be deprecated one day.
    if target_number:
        if isinstance(target_number[0], int):
            target_number = target_number[1], target_number[0]
        return StoppingCriteria(*target_number)

    return None


# Entry point to Snowfakery used by both the API ("snowfakery.generate_data")
# and the command line ("snowfakery.cli")
def generate_data(
    yaml_file: FileLike,
    *,
    parent_application: SnowfakeryApplication = None,  # the parent application
    user_options: T.Dict[str, T.Any] = None,  # same as --option
    dburl: str = None,  # same as --dburl
    dburls=[],  # same as multiple --dburl options
    target_number: T.Tuple = None,  # same as --target-number
    debug_internals: bool = None,  # same as --debug-internals
    generate_cci_mapping_file: FileLike = None,  # same as --generate-cci-mapping-file
    output_format: str = None,  # same as --output-format
    output_file: FileLike = None,  # same as --output-file
    output_files: T.List[FileLike] = None,  # same as multiple --output-file options
    output_folder: FileLike = None,  # same as --output-folder
    continuation_file: FileLike = None,  # continuation file from last execution
    generate_continuation_file: FileLike = None,  # place to generate continuation file
    should_create_cci_record_type_tables: bool = False,  # create CCI Record type tables?
    load_declarations: T.Sequence[
        FileLike
    ] = None,  # read these load declarations for CCI
    plugin_options: T.Mapping = None,
    update_input_file: FileLike = None,  # use this input file in update mode
    update_passthrough_fields: T.Sequence[
        str
    ] = (),  # pass through these fields from input to output
) -> None:
    stopping_criteria = stopping_criteria_from_target_number(target_number)
    dburls = dburls or ([dburl] if dburl else [])
    output_files = output_files or []
    if output_file:
        output_files = output_files + [output_file]

    with ExitStack() as exit_stack:

        def open_with_cleanup(file, mode, **kwargs):
            return exit_stack.enter_context(open_file_like(file, mode, **kwargs))

        parent_application = parent_application or SnowfakeryApplication(
            stopping_criteria
        )

        output_stream = exit_stack.enter_context(
            configure_output_stream(
                dburls, output_format, output_files, output_folder, parent_application
            )
        )

        yaml_path, open_yaml_file = open_with_cleanup(yaml_file, "r")
        _, open_new_continue_file = open_with_cleanup(generate_continuation_file, "w")
        _, open_continuation_file = open_with_cleanup(continuation_file, "r")
        _, open_cci_mapping_file = open_with_cleanup(generate_cci_mapping_file, "w")
        # utf-8-sig and newline="" are for Windows
        _, open_update_input_file = open_with_cleanup(
            update_input_file, "r", newline="", encoding="utf-8-sig"
        )

        summary = generate(
            open_yaml_file=open_yaml_file,
            user_options=user_options,
            output_stream=output_stream,
            parent_application=parent_application,
            generate_continuation_file=open_new_continue_file,
            continuation_file=open_continuation_file,
            stopping_criteria=stopping_criteria,
            plugin_options=plugin_options,
            update_input_file=open_update_input_file,
            update_passthrough_fields=update_passthrough_fields,
        )

        if open_cci_mapping_file:
            declarations = gather_declarations(yaml_path or "", load_declarations)
            yaml.safe_dump(
                mapping_from_recipe_templates(summary, declarations),
                open_cci_mapping_file,
                sort_keys=False,
            )
    if should_create_cci_record_type_tables:
        create_cci_record_type_tables(dburls[0])


@contextmanager
def configure_output_stream(
    dburls, output_format, output_files, output_folder, parent_application
):
    assert isinstance(output_files, (list, type(None)))

    with _get_output_streams(
        dburls, output_files, output_format, output_folder
    ) as output_streams:
        if len(output_streams) == 0:
            output_stream = DebugOutputStream()
        elif len(output_streams) == 1:
            output_stream = output_streams[0]
        else:
            output_stream = MultiplexOutputStream(output_streams)
        try:
            yield output_stream
        finally:
            try:
                messages = output_stream.close()
            except Exception as e:
                messages = None
                parent_application.echo(
                    f"Could not close {output_stream}: {str(e)}", err=True
                )
            if messages:
                for message in messages:
                    parent_application.echo(message)


@contextmanager
def _get_output_streams(dburls, output_files, output_format, output_folder):
    with ExitStack() as onexit:
        output_streams = []  # we allow multiple output streams
        for dburl in dburls:
            output_streams.append(SqlDbOutputStream.from_url(dburl))

        if output_format and not output_files:
            output_stream_cls = get_output_stream_class(output_format)

            if output_stream_cls.is_text and not output_files:
                output_streams.append(output_stream_cls(sys.stdout))

            if output_stream_cls.uses_folder:
                output_streams.append(output_stream_cls(output_folder))

        if output_files:
            for f in output_files:
                format = output_format
                if output_folder and isinstance(f, (str, Path)):
                    f = Path(output_folder, f)  # put the file in the output folder
                file_context = open_file_like(f, "w")
                path, open_file = onexit.enter_context(file_context)
                if path and not format:
                    format = Path(path).suffix[1:].lower()

                if not format:
                    raise exc.DataGenError("No format supplied or inferrable")

                output_stream_cls = get_output_stream_class(format)
                if output_stream_cls.uses_path:
                    open_file.close()
                    open_file = path
                output_streams.append(output_stream_cls(open_file, format=format))

        yield output_streams


def get_output_stream_class(output_format):
    from snowfakery.plugins import resolve_plugin_alternatives, plugin_path

    if "." in output_format:
        output_stream_classname = output_format
    else:
        output_format = output_format.lower() if output_format else None
        output_stream_classname = OUTPUT_FORMATS.get(output_format)
    if not output_stream_classname:
        raise exc.DataGenError(f"Unknown format or file extension: {output_format}")
    with plugin_path([]):
        rc = resolve_plugin_alternatives(output_stream_classname)
    if not rc:
        raise exc.DataGenError(f"Cannot load {output_stream_classname}")
    return rc


def gather_declarations(yaml_file, load_declarations):
    """Gather declarations from load declaration files."""
    if not load_declarations:
        inferred_load_file_path = infer_load_file_path(yaml_file)
        if inferred_load_file_path.is_file():
            load_declarations = [inferred_load_file_path]

    if load_declarations:
        declarations = []
        for declfile in load_declarations:
            with open_file_like(declfile, "r") as (path, f):
                decls = SObjectRuleDeclarationFile.parse_from_yaml(f)
                declarations.extend(decls.sobject_declarations)

        unified_declarations = unify(declarations)
    else:
        unified_declarations = {}
    return unified_declarations


def infer_load_file_path(yaml_file: T.Union[str, Path]):
    """Infer a load declaration from a filename"""
    yaml_file = str(yaml_file)
    suffixes = "".join(Path(yaml_file).suffixes)
    if suffixes:
        return Path(yaml_file.replace(suffixes, ".load.yml"))
    else:
        return Path("")



from snowfakery.plugins import SnowfakeryPlugin
from pathlib import Path


class File(SnowfakeryPlugin):
    class Functions:
        def file_data(self, file, encoding="utf-8"):
            if encoding == "binary":
                encoding = "latin-1"

            template_path = Path(self.context.current_filename).parent

            with open(template_path / file, "rb") as data:
                return data.read().decode(encoding)



from multiprocessing import Process, set_start_method
from threading import Thread
from time import time, sleep
from tempfile import TemporaryDirectory
from pathlib import Path
from collections import defaultdict
import locale

import click
from sqlalchemy import create_engine, inspect, text

from snowfakery import generate_data

benchmark_1 = Path(__file__).parent / "benchmark_1.yml"
try:
    locale.setlocale(locale.LC_ALL, "")
except locale.Error as e:
    print(str(e))


@click.command()
@click.argument("recipe", type=click.Path(), default=str(benchmark_1))
@click.option("--num-records", type=int, default=100_000)
@click.option("--num-records-tablename", type=str, default="Account")
@click.option("--number-of-processes", type=int, default=0)
@click.option("--number-of-threads", type=int, default=0)
def snowbench(
    num_records,
    num_records_tablename,
    recipe,
    number_of_processes=0,
    number_of_threads=0,
):
    """A benchmarking tool for Snowfakery and Snowfakery recipes.

    Note that Snowfakery itself is not multi-threaded or multi-process,
    but it takes only a few lines of code to wrap it in a parallel
    processing context. For example, it might be a few lines of
    shell script or 10 lines of Python code.

    The sweet spot for "number_of_processes" is usually the number of
    CPU cores you have. Processes are usually faster than threads for
    Snowfakery.

    In some weird circumstances number_of_processes=1 may behave
    differently than the default behaviour, because the single
    execution process is spawned off into a sub-process instead
    of being executed inline.
    """
    with TemporaryDirectory() as tempdir, click.progressbar(
        label="Benchmarking", length=num_records, show_eta=False
    ) as progress_bar:
        start = time()
        Thread(
            daemon=True,
            target=status,
            args=[
                tempdir,
                num_records,
                num_records_tablename,
                progress_bar,
            ],
        ).start()
        if not number_of_processes and not number_of_threads:
            print(
                "--number-of-processes and --number-of-threads not supplied.\n"
                "Using a single inline execution unit."
            )
            output_file = Path(tempdir) / "generated_data.db"
            snowfakery(recipe, num_records, num_records_tablename, output_file)
        else:
            multithreaded_test(
                recipe,
                num_records,
                num_records_tablename,
                tempdir,
                number_of_processes,
                number_of_threads,
            )
        progress_bar.update(num_records)
        report_databases(tempdir, start, num_records_tablename)


def multithreaded_test(
    recipe,
    num_records,
    num_records_tablename,
    tempdir,
    number_of_processes,
    number_of_threads,
):
    num_workers = number_of_processes + number_of_threads

    def make_args(tempdir, prefix, idx):
        output_file = f"{tempdir}/sqlite_{prefix}_{idx}.db"
        return recipe, num_records // num_workers, num_records_tablename, output_file

    processes = [
        Process(target=snowfakery, args=make_args(tempdir, "p", idx))
        for idx in range(number_of_processes)
    ]
    threads = [
        Thread(target=snowfakery, args=make_args(tempdir, "t", idx))
        for idx in range(number_of_threads)
    ]

    workers = threads + processes
    start = time()
    for worker in workers:
        worker.start()

    for worker in workers:
        worker.join()

    duration = time() - start
    return duration


def count_databases(tempdir):
    databases = Path(tempdir).glob("*.db")
    counts = defaultdict(int)
    for database in databases:
        count_database(database, counts)
    return dict(counts)


def status(tempdir, num_records, num_records_tablename, progress_bar):
    start = time()
    sleep(2)
    previous_count = 0
    for i in range(1, 10**20):
        sleep(1)
        if i in (2, 5, 10, 20, 30, 45, 90, 150) or (i % 60 == 0):
            print()
            current_count = report_databases(tempdir, start, num_records_tablename)
            progress_bar.update(current_count - previous_count)
            previous_count = current_count


def report_databases(tempdir, start_time, relevant_table_name):
    print()
    print()
    counts = count_databases(tempdir)
    for name, count in counts.items():
        print(name, f"{count:n}")
    total = sum(counts.values())
    duration = time() - start_time
    print(
        f"{total:n} records /",
        f"{int(duration):n} seconds =~",
        f"{int(total / duration):n}",
        "records per second",
        f"\n= {int((total / duration) * 3600):n}",
        "records per hour",
        f"\n= {int((total / duration) * 3600 *24):n}",
        "records per day",
    )
    return counts[relevant_table_name]


def count_database(filename, counts):
    dburl = f"sqlite:///{filename}?mode=ro"
    engine = create_engine(dburl)
    insp = inspect(engine)
    tables = insp.get_table_names()  # type: ignore
    for table in tables:
        counts[table] += count_table(engine, table)
    return counts


def count_table(engine, tablename):
    with engine.connect() as c:
        return c.execute(text(f"select count(Id) from '{tablename}'")).first()[0]


def snowfakery(recipe, num_records, tablename, outputfile):
    assert Path(recipe).exists(), recipe
    output = f"sqlite:///{outputfile}"
    print(output)
    generate_data(
        recipe,
        target_number=(num_records, tablename),
        dburl=output,
    )


def main():
    snowbench.main(prog_name="snowbench")


if __name__ == "__main__":  # pragma: no cover
    set_start_method("spawn")
    main()



import typing as T
from pathlib import Path

if T.TYPE_CHECKING:
    from contextlib import contextmanager
else:
    # until https://github.com/agronholm/typeguard/issues/115 is fixed in typeguard 3
    from snowfakery.backports.typeguard_context_manager_hack import contextmanager

from click.utils import LazyFile

OpenFileLike = T.Union[T.TextIO, LazyFile]
FileLike = T.Union[OpenFileLike, Path, str]


@contextmanager
def open_file_like(
    file_like: T.Optional[FileLike], mode, **kwargs
) -> T.Generator[T.Tuple[T.Optional[Path], T.Optional[OpenFileLike]], None, None]:
    """Look at a file-like or path-like object and open it

    Returns a) the path it was given OR a best-guess of the path and
            b) the open file object

    Used properly as a context manager, it will close the file later
    if appropriate.
    """
    if not file_like:
        yield None, None
    if isinstance(file_like, str):
        file_like = Path(file_like)

    if isinstance(file_like, Path):
        with file_like.open(mode, **kwargs) as f:
            yield file_like, f

    elif hasattr(file_like, "name"):
        yield Path(file_like.name), file_like  # type: ignore

    elif hasattr(file_like, "read"):
        yield None, T.cast(OpenFileLike, file_like)



import random
import typing as T
import datetime
from difflib import get_close_matches
from itertools import product

import dateutil
from faker import Faker, Generator

from snowfakery.plugins import PluginContext
import snowfakery.data_gen_exceptions as exc

# .format language doesn't allow slicing. :(
first_name_patterns = ("{firstname}", "{firstname[0]}", "{firstname[0]}{firstname[1]}")
first_name_separators = ("", ".", "-", "_", "+")
year_patterns = ("{year}", "{year[2]}{year[3]}", "{year[3]}", "")

email_templates = [
    f"{first_name}{first_name_separator}{{lastname}}{year}@{{domain}}"
    for first_name, first_name_separator, year in product(
        first_name_patterns, first_name_separators, year_patterns
    )
]

this_year = datetime.datetime.today().year
DateLike = T.Union[datetime.date, datetime.datetime, datetime.timedelta, str, int]
TimeZoneAsRelDelta = T.Union[dateutil.relativedelta.relativedelta, T.Literal[False]]
UTCAsRelDelta = dateutil.relativedelta.relativedelta(hours=0)


class FakeNames(T.NamedTuple):
    f: Faker
    faker_context: PluginContext = None

    # "matching" allows us to turn off the behaviour of
    # trying to incorporate one field into another if we
    # need to.
    def user_name(self, matching: bool = True):
        "Globally unique Salesforce-style username in the form of an email address"
        domain = self.f.hostname()
        already_created = self._already_have(("firstname", "lastname"))
        if matching and all(already_created):
            namepart = f"{already_created[0]}.{already_created[1]}_{self.f.uuid4()}"
        else:
            namepart = f"{self.f.first_name()}_{self.f.last_name()}_{self.f.uuid4()}"

        namepart_max_len = 80 - (len(domain) + 1)
        namepart = namepart[0:namepart_max_len]
        return f"{namepart}@{domain}"

    def alias(self):
        """Salesforce-style 8-character alias: really an 8 char-truncated firstname.
        Not necessarily unique, but likely to be unique if you create small
        numbers of them."""
        return self.f.first_name()[0:8]

    def email(self, matching: bool = True):
        """Email address using one of the "example" domains"""
        already_created = self._already_have(("firstname", "lastname"))
        if matching and all(already_created):
            template = random.choice(email_templates)

            return template.format(
                firstname=already_created[0].ljust(2, "_"),
                lastname=already_created[1],
                domain=self.f.safe_domain_name(),
                year=str(random.randint(this_year - 80, this_year - 10)),
            )
        return self.f.ascii_safe_email()

    def realistic_maybe_real_email(self):
        """Like fake: email except that the email domain may be real and therefore
        the email address itself may be real. Use with caution, you might
        accidentally email strangers!!!
        """
        return self.f.email()

    def _already_have(self, names: T.Sequence[str]):
        """Get a list of field values that we've already generated"""
        already_created = self.faker_context.local_vars()
        vals = [already_created.get(name) for name in names]

        # if we ever need to use this function in a context where Unicode
        # is okay, here's how to implement that:
        # cleanup_func = replace_unicode_strings_with_None if must_be_ascii else NOOP

        vals = [replace_unicode_strings_with_None(val) for val in vals]

        return vals

    def state(self):
        """Return a state, province or other appropriate administrative unit"""
        return self.f.administrative_unit()

    def postalcode(self):
        """Return whatever counts as a postalcode for a particular locale"""
        return self.f.postcode()

    def date_time_between(
        self,
        start_date: DateLike = "-30y",
        end_date: DateLike = "now",
        timezone: TimeZoneAsRelDelta = UTCAsRelDelta,
    ) -> datetime.datetime:
        timezone = _normalize_timezone(timezone)
        return self.f.date_time_between(start_date, end_date, timezone)

    date_time_between_dates = date_time_between

    def future_datetime(
        self,
        end_date: DateLike = "+30d",
        timezone: TimeZoneAsRelDelta = UTCAsRelDelta,
    ) -> datetime.datetime:
        timezone = _normalize_timezone(timezone)
        return self.f.future_datetime(end_date, timezone)

    def iso8601(
        self,
        timezone: TimeZoneAsRelDelta = UTCAsRelDelta,
        end_datetime: DateLike = None,
    ) -> str:
        timezone = _normalize_timezone(timezone)
        return self.f.iso8601(
            timezone,
            end_datetime,
        )

    date_time = datetime = date_time_between

    # These faker types are not available in Snowfakery
    # because they are redundant
    date_time_this_year = NotImplemented
    date_time_this_year = NotImplemented
    date_time_this_month = NotImplemented
    date_time_ad = NotImplemented
    date_time_this_century = NotImplemented
    date_time_this_decade = NotImplemented


def _normalize_timezone(timezone=None):
    if timezone in (None, False):
        return None
    elif timezone is UTCAsRelDelta:
        return datetime.timezone.utc
    else:
        if not isinstance(timezone, dateutil.relativedelta.relativedelta):
            raise exc.DataGenError(  # pragma: no cover
                f"`timezone` should be a `relativedelta`, not `{type(timezone).__name__}`: {timezone}"
            )
        return datetime.timezone(
            datetime.timedelta(hours=timezone.hours, minutes=timezone.minutes)
        )


# we will use this to exclude Faker's internal book-keeping methods
# from our faker interface
faker_class_attrs = set(dir(Faker)).union((dir(Generator)))


class FakeData:
    """Wrapper for Faker which adds Salesforce names and case insensitivity."""

    def __init__(
        self,
        faker_providers: T.Sequence[object],
        locale: T.Optional[str] = None,
        faker_context: T.Optional[PluginContext] = None,
    ):
        # access to persistent state
        self.faker_context = faker_context

        faker = Faker(locale, use_weighting=False)
        for provider in faker_providers:
            faker.add_provider(provider)

        fake_names = FakeNames(faker, faker_context)

        def no_underscore_name(name):
            return name.lower().replace("_", "")

        def obj_to_func_list(obj: object, canonicalizer: T.Callable, ignore_list: set):
            return {
                canonicalizer(name): getattr(obj, name)
                for name in dir(obj)
                if not name.startswith("_") and name not in ignore_list
            }

        # canonical form of names is lower-case, no underscores
        # include faker names with underscores in case of ab_c/a_bc clashes
        # include faker names with no underscores to emulate salesforce
        # include snowfakery names defined above
        self.fake_names = {
            **obj_to_func_list(faker, str.lower, faker_class_attrs),
            **obj_to_func_list(faker, no_underscore_name, faker_class_attrs),
            # in case of conflict, snowfakery names "win" over Faker names
            **obj_to_func_list(fake_names, str.lower, set()),
            **obj_to_func_list(fake_names, no_underscore_name, set()),
        }

    def _get_fake_data(self, origname, *args, **kwargs):
        local_faker_vars = self.faker_context.local_vars()

        # faker names are all lower-case
        name = origname.lower()

        meth = self.fake_names.get(name, NotImplemented)

        if meth != NotImplemented:
            ret = meth(*args, **kwargs)
            local_faker_vars[name.replace("_", "")] = ret
            return ret

        msg = f"No fake data type named {origname}."
        all_fake_names = [k for k, v in self.fake_names.items() if v != NotImplemented]
        match_list = get_close_matches(name, all_fake_names, n=1)
        if match_list:
            msg += f" Did you mean {match_list[0]}"
        raise AttributeError(msg)


def translate(x):
    if chr(x).isalnum():
        return x
    else:
        return None


REMOVE_WEIRD_CHARS = {x: translate(x) for x in range(0, 128)}


def replace_unicode_strings_with_None(val):
    if type(val) == str:
        if not val.isascii():
            return None
        return val.translate(REMOVE_WEIRD_CHARS)
    return val



import typing as T
from collections import defaultdict


from snowfakery.data_generator import ExecutionSummary
from snowfakery.salesforce import find_record_type_column
from snowfakery.data_generator_runtime import Dependency
from snowfakery.utils.collections import OrderedSet
from snowfakery.parse_recipe_yaml import TableInfo

from .cci_mapping_files.declaration_parser import SObjectRuleDeclaration
from .cci_mapping_files.post_processes import add_after_statements


class LoadStep(T.NamedTuple):
    action: str
    table_name: str
    update_key: T.Optional[str]
    fields: T.Sequence[str]


def mapping_from_recipe_templates(
    summary: ExecutionSummary,
    declarations: T.Mapping[str, SObjectRuleDeclaration] = None,
):
    """Use the outputs of the recipe YAML and convert to Mapping.yml format"""
    declarations = declarations or {}
    relevant_declarations = [decl for decl in declarations.values() if decl.load_after]

    inferred_dependencies, declared_dependencies, reference_fields = build_dependencies(
        summary.intertable_dependencies, relevant_declarations
    )
    tables = summary.tables.copy()
    remove_person_contact_id(inferred_dependencies, tables)
    table_order = sort_dependencies(
        inferred_dependencies, declared_dependencies, tables
    )
    load_steps = load_steps_from_tableinfos(tables, table_order)
    mappings = mappings_from_load_steps(load_steps, reference_fields, declarations)
    return mappings


def load_steps_from_tableinfos(
    tables: T.Dict[str, TableInfo], table_order: T.List[str]
) -> T.List[LoadStep]:
    load_steps = OrderedSet()
    for table_name, tableinfo in tables.items():
        for template in tableinfo._templates:

            if template.update_key:
                action = "upsert"
            else:
                action = "insert"

            load_steps.add(
                LoadStep(
                    action,
                    table_name,
                    template.update_key,
                    tuple(tableinfo.fields.keys()),
                )
            )

    load_steps_as_list = list(load_steps)
    load_steps_as_list.sort(key=lambda step: table_order.index(step.table_name))
    return load_steps_as_list


def remove_person_contact_id(dependencies, tables):
    if "Account" in dependencies:
        dep_to_person_contact = [
            dep
            for dep in dependencies["Account"]
            if dep.table_name_to.lower() == "personcontact"
        ]
        for dep in dep_to_person_contact:
            dependencies["Account"].remove(dep)

    if tables.get("Account") and tables["Account"].fields.get("PersonContactId"):
        del tables["Account"].fields["PersonContactId"]


def build_dependencies(
    intertable_dependencies,
    declarations: T.Sequence[SObjectRuleDeclaration] = None,
):
    """Figure out which tables depend on which other ones (through foreign keys)

    intertable_dependencies is a dict with values of Dependency objects.

    Returns two things:
        1. a dictionary allowing easy lookup of dependencies by parent table
        2. a dictionary allowing lookups by (tablename, fieldname) pairs
    """
    inferred_dependencies = defaultdict(OrderedSet)
    declared_dependencies = defaultdict(OrderedSet)
    reference_fields = {}
    declarations = declarations or ()

    for dep in intertable_dependencies:
        table_deps = inferred_dependencies[dep.table_name_from]
        table_deps.add(dep)
        reference_fields[(dep.table_name_from, dep.field_name)] = dep.table_name_to
    for decl in declarations:
        assert isinstance(decl.load_after, list)
        for target in decl.load_after:
            declared_dependencies[decl.sf_object].add(
                Dependency(decl.sf_object, target, "(none)")
            )

    return inferred_dependencies, declared_dependencies, reference_fields


def _table_is_free(table_name, dependencies, sorted_tables):
    """Check if every child of this table is already sorted

    Look at the unit test test_table_is_free_simple to see some
    usage examples.
    """
    tables_this_table_depends_upon = dependencies.get(table_name, OrderedSet()).copy()
    for dependency in sorted(tables_this_table_depends_upon):
        if (
            dependency.table_name_to in sorted_tables
            or dependency.table_name_to == table_name
        ):
            tables_this_table_depends_upon.remove(dependency)

    return len(tables_this_table_depends_upon) == 0


def sort_dependencies(inferred_dependencies, declared_dependencies, tables):
    """Sort the dependencies to output tables in the right order."""
    dependencies = {**inferred_dependencies, **declared_dependencies}
    sorted_tables = []

    while tables:
        remaining = len(tables)
        leaf_tables = [
            table
            for table in tables
            if _table_is_free(table, dependencies, sorted_tables)
        ]
        sorted_tables.extend(leaf_tables)
        tables = [table for table in tables if table not in sorted_tables]
        if len(tables) == remaining:

            # this is a bit tricky.
            # run the algorithm with ONLY the declared
            # dependencies and see if it comes to resolution
            if inferred_dependencies and declared_dependencies:
                subset = sort_dependencies({}, declared_dependencies, tables.copy())
                sorted_tables.extend(subset)
            else:
                sorted_tables.append(sorted(tables)[0])

    return sorted_tables


def mappings_from_load_steps(
    load_steps: T.List[LoadStep],
    reference_fields: dict,
    declarations: T.Mapping[str, SObjectRuleDeclaration],
):
    """Generate mapping.yml data structures."""
    mappings = {}
    for load_step in load_steps:
        table_name = load_step.table_name
        record_type_col = find_record_type_column(table_name, load_step.fields)

        fields = {
            fieldname: fieldname
            for fieldname in load_step.fields
            if (table_name, fieldname) not in reference_fields.keys()
            and fieldname != record_type_col
        }
        if record_type_col:
            fields["RecordTypeId"] = record_type_col

        lookups = {
            fieldname: {
                "table": reference_fields[(table_name, fieldname)],
                "key_field": fieldname,
            }
            for fieldname in load_step.fields
            if (table_name, fieldname) in reference_fields.keys()
        }
        if table_name == "PersonContact":
            sf_object = "Contact"
        else:
            sf_object = table_name
        mapping = {"sf_object": sf_object, "table": table_name, "fields": fields}
        if lookups:
            mapping["lookups"] = lookups

        sobject_declarations = declarations.get(table_name)
        if sobject_declarations:
            mapping.update(sobject_declarations.as_mapping())

        if load_step.update_key:
            mapping["action"] = "upsert"
            mapping["update_key"] = load_step.update_key
            mapping["filters"] = [f"_sf_update_key = '{load_step.update_key}'"]
            step_name = f"Upsert {table_name} on {load_step.update_key}"
        else:
            step_name = f"Insert {table_name}"
            any_other_loadstep_for_this_table_has_update_key = any(
                ls
                for ls in load_steps
                if (ls.table_name == table_name and ls.update_key)
            )
            if any_other_loadstep_for_this_table_has_update_key:
                mapping["filters"] = ["_sf_update_key = NULL"]

        mappings[step_name] = mapping

    add_after_statements(mappings)
    return mappings



#!/usr/bin/env python3
import sys
from pathlib import Path

from snowfakery.data_gen_exceptions import DataGenError
from snowfakery.utils.versions import check_latest_version

import click
from snowfakery.__about__ import __version__ as version
from snowfakery.api import file_extensions, generate_data, COUNT_REPS

if __name__ == "__main__":  # pragma: no cover
    sys.path.append(str(Path(__file__).parent.parent))


class FormatChoice(click.Choice):
    def convert(self, value, param, ctx):
        if isinstance(value, str) and "." in value:
            return value

        return super().convert(value, param, ctx)


def eval_arg(arg):
    if arg.isnumeric():
        return int(arg)
    else:
        return arg


# don't add a type signature to this function.
# typeguard and click will interfere with each other.
def int_string_tuple(ctx, param, value=None):
    """
    Parse a pair of strings that represent a string and a number.

    Either number, string or string, number is allowed as input."""
    if not value:
        return None
    assert len(value) == 2

    try:
        number, string = int(value[0]), value[1]
    except ValueError:
        try:
            string, number = value[0], int(value[1])
        except ValueError:
            raise click.BadParameter(
                "This parameter must be of the form 'number Name'. For example '50 Account'"
            )
    return string, number


class VersionMessage:
    "Class that formats itself as a version message when % interpolated"

    def quick_test(self) -> str:
        from io import StringIO

        yaml = """
        - object: Status
          fields:
            quote: Shiny and new
        """
        out = StringIO()
        generate_data(StringIO(yaml), output_files=[out], output_format="txt")
        return "Properly installed" if out.getvalue() else "Unknown installation error"

    def __mod__(self, vals) -> str:
        return "\n".join(
            (
                f"snowfakery version {version}",
                check_latest_version(version).message,
                "",
                f"Program: {__file__}",
                f"Python: {sys.version}",
                f"Executable: {sys.executable}",
                f"Installation: {self.quick_test()}",
            )
        )


@click.command()
@click.argument("yaml_file", type=click.File("r"))
@click.option(
    "--dburl",
    "dburls",
    type=str,
    multiple=True,
    help="URL for database to save data to. "
    "Use sqlite:///foo.db if you don't have one set up.",
)
@click.option(
    "--output-format",
    "output_format",
    type=FormatChoice(file_extensions, case_sensitive=False),
)
@click.option("--output-folder", "output_folder", type=click.Path(), default=".")
@click.option("--output-file", "-o", "output_files", type=click.Path(), multiple=True)
@click.option(
    "--option",
    nargs=2,
    type=eval_arg,
    multiple=True,
    help="Option to send to the recipe YAML in a format like 'OptName OptValue'. Specify multiple times if needed.",
)
@click.option(
    "--target-number",
    "--target-count",
    nargs=2,
    help="Target record count for the recipe YAML in the form of 'number tablename'. "
    "For example: '50 Account' to generate roughly 50 accounts.",
    callback=int_string_tuple,  # noqa  https://github.com/pallets/click/issues/789#issuecomment-535121714
)
@click.option(
    "--reps",
    help="Target repetition count for the recipe YAML. Use as an alternative to --target-number",
    type=int,
)
@click.option(
    "--debug-internals/--no-debug-internals", "debug_internals", default=False
)
@click.option(
    "--generate-cci-mapping-file",
    type=click.File("w"),
    help="Generate a CumulusCI mapping file for the dataset",
)
@click.option(
    "--generate-continuation-file",
    type=click.File("w"),
    help="A file that captures information about how to continue a "
    "multi-batch data generation process",
)
@click.option(
    "--generate-record-type-tables/--no-generate-record-type-tables",
    "should_create_cci_record_type_tables",
    default=False,
    hidden=True,
)
@click.option(
    "--continuation-file",
    type=click.File("r"),
    help="Continue generating a dataset where 'continuation-file' left off",
)
@click.option(
    "--plugin-option",
    nargs=2,
    type=eval_arg,
    multiple=True,
    help="Option to send to a plugin in a format like 'OptName OptValue'. Specify multiple times if needed.",
)  # options passed by an API instead of CLI
@click.option(
    "--load-declarations",
    type=click.Path(exists=True, readable=True, dir_okay=False),
    help="Declarations to mix into the generated mapping file",
    multiple=True,
)
@click.option(
    "--update-input-file",
    type=click.Path(exists=True, readable=True, dir_okay=False),
    help="Run an update-style recipe on this input CSV",
)
@click.option(
    "--update-passthrough-fields",
    type=str,
    help="Pass through these fields from input to output automatically",
    hidden=True,
    default=None,
)
@click.version_option(version=version, prog_name="snowfakery", message=VersionMessage())
def generate_cli(
    yaml_file,
    option=(),
    dburls=(),
    target_number=None,
    reps=None,
    debug_internals=None,
    generate_cci_mapping_file=None,
    output_format=None,
    output_files=None,
    output_folder=None,
    continuation_file=None,
    generate_continuation_file=None,
    plugin_option=(),
    should_create_cci_record_type_tables=False,
    load_declarations=None,
    update_input_file=None,
    update_passthrough_fields=(),  # undocumented feature used mostly for testing
):
    """
        Generates records from a YAML file

    \b
        Records can go to:
            * stdout (default)
            * JSON file (--output-format=json --output-file=foo.json)
            * diagram file (--output-format=png --output-file=foo.png)
            * a database identified by --dburl (e.g. --dburl sqlite:////tmp/foo.db)
            * or to a directory as a set of CSV files (--output-format=csv --output-folder=csvfiles)

        Diagram output depends on the installation of graphviz (https://www.graphviz.org/download/)

        Full documentation here:

            * https://snowfakery.readthedocs.io/en/docs/
    """
    output_files = list(output_files) if output_files else []
    validate_options(
        yaml_file=yaml_file,
        option=option,
        dburl=dburls,
        debug_internals=debug_internals,
        generate_cci_mapping_file=generate_cci_mapping_file,
        output_format=output_format,
        output_files=output_files,
        output_folder=output_folder,
        target_number=target_number,
        reps=reps,
    )
    if update_passthrough_fields:
        update_passthrough_fields = update_passthrough_fields.split(",")
    else:
        update_passthrough_fields = ()
    try:
        user_options = dict(option)
        plugin_options = dict(plugin_option)
        if reps:
            target_number = (COUNT_REPS, reps)

        generate_data(
            yaml_file=yaml_file,
            user_options=user_options,
            dburls=dburls,
            target_number=target_number,
            debug_internals=debug_internals,
            generate_cci_mapping_file=generate_cci_mapping_file,
            output_format=output_format,
            output_files=output_files,
            output_folder=output_folder,
            continuation_file=continuation_file,
            generate_continuation_file=generate_continuation_file,
            should_create_cci_record_type_tables=should_create_cci_record_type_tables,
            load_declarations=load_declarations,
            plugin_options=plugin_options,
            update_input_file=update_input_file,
            update_passthrough_fields=update_passthrough_fields,
        )
    except DataGenError as e:
        if debug_internals:
            raise e
        else:
            click.echo("")
            click.echo(e.prefix)
            raise click.ClickException(str(e)) from e


def validate_options(
    yaml_file,
    option,
    dburl,
    debug_internals,
    generate_cci_mapping_file,
    output_format,
    output_files,
    output_folder,
    target_number,
    reps,
):
    if dburl and output_format:
        raise click.ClickException(
            "Sorry, you need to pick --dburl or --output-format "
            "because they are mutually exclusive."
        )
    if dburl and output_files:
        raise click.ClickException(
            "Sorry, you need to pick --dburl or --output-file "
            "because they are mutually exclusive."
        )
    if (
        output_folder
        and str(output_folder) != "."
        and not (output_files or output_format == "csv")
    ):
        raise click.ClickException(
            "--output-folder can only be used with --output-file=<something> or --output-format=csv"
        )

    if target_number and reps:
        raise click.ClickException(
            "Sorry, you need to pick --target_number or --reps "
            "because they are mutually exclusive."
        )


def main():
    generate_cli.main(prog_name="snowfakery")


if __name__ == "__main__":  # pragma: no cover
    main()



import os
from contextlib import contextmanager, ExitStack
from csv import DictReader
from pathlib import Path
from random import shuffle
from typing import Any, Optional

from sqlalchemy import MetaData, create_engine
from sqlalchemy.sql.elements import quoted_name
from sqlalchemy.sql.expression import func, select
from yaml.representer import Representer

from snowfakery.data_gen_exceptions import DataGenError, DataGenNameError
from snowfakery.plugins import (
    PluginResult,
    PluginResultIterator,
    SnowfakeryPlugin,
    memorable,
)
from snowfakery.utils.files import FileLike, open_file_like
from snowfakery.utils.yaml_utils import SnowfakeryDumper


def _open_db(db_url):
    "Internal function for opening the database up."
    engine = create_engine(db_url)
    metadata = MetaData()
    metadata.reflect(views=True, bind=engine)
    return engine, metadata


def sql_dataset(
    db_url: str, tablename: Optional[str] = None, mode="linear", repeat: bool = True
):
    "Open the right SQL Dataset iterator based on the params"
    assert db_url
    engine, metadata = _open_db(db_url)
    tables = {
        name: value
        for name, value in metadata.tables.items()
        if not name.startswith("sqlite")
    }
    table = None
    if tablename:
        table = tables.get(tablename)
        if table is None:
            raise AttributeError(f"Cannot find table: {tablename}")
    elif len(tables) == 0:
        raise Exception("Database does not exist or has no tables in it")
    elif len(tables) == 1:
        table = next(iter(tables.values()))
    elif len(tables) > 1:
        raise Exception(
            f"Database has multiple tables in it and none was selected: {metadata.tables.keys()}"
        )
    if mode == "linear":
        return SQLDatasetLinearIterator(engine, table, repeat)
    elif mode == "shuffle":
        return SQLDatasetRandomPermutationIterator(engine, table, repeat)
    raise AssertionError(f"Unknown mode: {mode}")


class DatasetIteratorBase(PluginResultIterator):
    """Base class for Dataset Iterators

    Subclasses should implement 'self.restart' which puts an iterator into 'self.results'
    """

    def __init__(self, repeat):
        # subclasses can register stuff to be cleaned up here.
        self.cleanup = ExitStack()
        super().__init__(repeat)

    def next_result(self):
        return next(self.results)

    def close(self):
        self.cleanup.close()


class SQLDatasetIterator(DatasetIteratorBase):
    def __init__(self, engine, table, repeat):
        self.connection = engine.connect()
        self.table = table
        super().__init__(repeat)
        self.start()

    def start(self):
        self.results = (
            DatasetPluginResult(dict(row._mapping))
            for row in self.connection.execute(self.query())
        )

    def close(self):
        self.results = None
        self.connection.close()
        super().close()

    def query(self):
        "Return a SQL Alchemy SELECT statement"
        raise NotImplementedError(f"query method on {self.__class__.__name__}")


class SQLDatasetLinearIterator(SQLDatasetIterator):
    "Iterator that reads a SQL table from top to bottom"

    def query(self):
        return select(self.table)


class SQLDatasetRandomPermutationIterator(SQLDatasetIterator):
    "Iterator that reads a SQL table in random order"

    def query(self):
        return select(self.table).order_by(func.random())


class CSVDatasetLinearIterator(DatasetIteratorBase):
    def __init__(self, datasource: FileLike, repeat: bool):
        super().__init__(repeat)
        # utf-8-sig and newline="" are for Windows
        self.path, self.file = self.cleanup.enter_context(
            open_file_like(datasource, "r", newline="", encoding="utf-8-sig")
        )

        self.start()

    def start(self):
        assert self.file
        self.file.seek(0)
        d = DictReader(self.file)  # type: ignore

        plugin_result = self.plugin_result
        self.results = (plugin_result(row) for row in d)

    def close(self):
        self.results = None
        super().close()

    def plugin_result(self, row):
        if None in row:
            raise DataGenError(
                f"Your CSV row has more columns than the CSV header:  {row[None]}, {self.path} {self.file}"
            )

        return DatasetPluginResult(row)


class DatasetPluginResult(PluginResult):
    def __getattr__(self, name):
        try:
            return super().__getattr__(name)
        except KeyError:
            raise DataGenNameError(
                f"`{name}` attribute not found. Should be one of {tuple(self.result.keys())}"
            )


class CSVDatasetRandomPermutationIterator(CSVDatasetLinearIterator):
    # This algorithm shuffles a million records without a problem on my laptop.
    # If you needed to scale it up to 40 or 50 times the scale, you could do
    # this instead:
    #   * don't read the whole file into memory. Just figure out where the line
    #     breaks are and shuffle the address of THOSE. Then seek to lines
    #     during parsing
    #
    # To scale even further:
    #
    #   * load the rows or indexes into a SQLite DB. Ask SQlite to generate
    #     another table that randomizes the rows. (haven't decided whether
    #     copying the rows up-front is better)
    #
    #   * segment the file into hundred-thousand-row partitions. Shuffle the
    #     rows in each partition and then pick randomly among the partitions
    #     before grabbing a row
    def start(self):
        assert self.file
        self.file.seek(0)
        d = DictReader(self.file)  # type: ignore
        rows = [DatasetPluginResult(row) for row in d]
        shuffle(rows)

        self.results = iter(rows)

    def close(self):
        self.results = None


class DatasetBase:
    def __init__(self, *args, **kwargs):
        self.datasets = {}

    def _get_dataset_instance(self, plugin_context, iteration_mode, kwargs):
        filename = plugin_context.field_vars()["template"].filename
        assert filename
        rootpath = Path(filename).parent
        dataset_instance = self._load_dataset(iteration_mode, rootpath, kwargs)
        return dataset_instance

    def _load_dataset(self, iteration_mode, rootpath, kwargs):
        raise NotImplementedError("_load_dataset not implemented")

    def close(self):
        raise NotImplementedError("close not implemented: " + repr(self))


class FileDataset(DatasetBase):
    def close(self):
        pass

    def _load_dataset(self, iteration_mode, rootpath, kwargs):
        dataset = kwargs.get("dataset")
        tablename = kwargs.get("table")
        repeat = kwargs.get("repeat", True)

        with chdir(rootpath):
            if "://" in dataset:
                return sql_dataset(dataset, tablename, iteration_mode, repeat)
            else:
                filename = Path(dataset)

                if not filename.exists():
                    raise FileNotFoundError("File not found:" + str(filename))

                if filename.suffix != ".csv":
                    raise AssertionError(
                        f"Filename extension must be .csv, not {filename.suffix}"
                    )

                if iteration_mode == "linear":
                    return CSVDatasetLinearIterator(filename, repeat)
                elif iteration_mode == "shuffle":
                    return CSVDatasetRandomPermutationIterator(filename, repeat)


class DatasetPluginBase(SnowfakeryPlugin):
    class Functions:
        context: Any

        @memorable
        def iterate(self, **kwargs):
            return self.context.plugin.dataset_impl._get_dataset_instance(
                self.context, "linear", kwargs
            )

        @memorable
        def shuffle(self, **kwargs):
            return self.context.plugin.dataset_impl._get_dataset_instance(
                self.context, "shuffle", kwargs
            )

    def close(self):
        if self.dataset_impl:
            self.dataset_impl.close()
            self.dataset_impl = None


class Dataset(DatasetPluginBase):
    def __init__(self, *args, **kwargs):
        self.dataset_impl = FileDataset()
        super().__init__(*args, **kwargs)


@contextmanager
def chdir(path):
    """Context manager that changes to another directory

    Not thread-safe!!!
    """
    cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(cwd)


SnowfakeryDumper.add_representer(quoted_name, Representer.represent_str)  # type: ignore



from random import Random, randint
from copy import copy
from math import log
from functools import lru_cache


@lru_cache()
def randomizer(key):
    return Random(key)


@lru_cache()
def mask_for_key(key, numbits):
    r = copy(randomizer(key))
    return r.getrandbits(numbits)


SHIFT1 = 10
SHIFT2 = SHIFT1 * 100
SHIFT3 = SHIFT2 * SHIFT1


def scramble_number(number: int, minbits: int = 10) -> int:
    "Make a number that looks random"

    assert minbits >= 10

    minbits = max(10, minbits - 13)  # I'm going to add about 13 bits at the end :(
    key = number % SHIFT1  # last digit (or bits) is a randomization key
    # remove the key from the number
    number = number // SHIFT1
    # how many bits I need to fiddle with
    numbits = max(minbits, (int(log(number, 2)) + 1) if number else minbits)
    assert numbits < SHIFT2
    # make a randomized mask based on the key
    mask = mask_for_key(key, numbits)
    # make the number look random by applying the mask
    scrambled = number ^ mask
    # return the number, the key and the number of bits that can reverse the scrambling
    return scrambled * SHIFT3 + key * SHIFT2 + numbits


def unscramble_number(number):
    # extract the numbits (least significant digits)
    numbits = number % SHIFT2
    # clear them out
    number = number - numbits
    # extract the key (next least significant)
    key = int((number % SHIFT3) / SHIFT2)
    # clear them out
    number = number - (key * SHIFT2)
    assert number % SHIFT3 == 0
    scrambled = number // SHIFT3
    # find the right mask
    mask = mask_for_key(key, numbits)
    unscrambled = scrambled ^ mask
    # put the key back
    return unscrambled * SHIFT1 + key


# If a function can be undone, then it is bijective
# i.e. each output has ONLY ONE input
def _test_scrambling_is_safe(iterations):
    for _ in range(0, iterations):
        for numsize in range(5, 100):
            for minbits in range(10, 100):
                num = randint(10, 10**numsize)
                scrambled = scramble_number(num, minbits)
                unscrambled = unscramble_number(scrambled)
                assert unscrambled == num, (num, minbits, scrambled)


if __name__ == "__main__":  # pragma: no cover
    _test_scrambling_is_safe(50)



import re
import pkg_resources
import requests
import typing as T


from json import JSONDecodeError

# This code is adapted from cumulusci/cli/utils.py
FINAL_VERSION_RE = re.compile(r"^[\d\.]+$")


def is_final_release(version: str) -> bool:
    """Returns bool whether version string should be considered a final release.

    Snowfakery versions are considered final if they contain only digits and periods.
    e.g. 1.0.1 is final but 2.0b1 and 2.0.dev0 are not.
    """
    return bool(FINAL_VERSION_RE.match(version))


def get_latest_final_version():
    """return the latest version of snowfakery in pypi, be defensive"""
    # use the pypi json api https://wiki.python.org/moin/PyPIJSON
    res = safe_json_from_response(
        requests.get("https://pypi.org/pypi/snowfakery/json", timeout=5)
    )
    versions = []
    for versionstring in res["releases"].keys():
        if not is_final_release(versionstring):
            continue
        versions.append(pkg_resources.parse_version(versionstring))
    versions.sort(reverse=True)
    return versions[0]


class IsLatestVersion(T.NamedTuple):
    is_latest_version: T.Optional[bool]
    message: str


def check_latest_version(version: str) -> IsLatestVersion:
    """checks for the latest version of snowfakery from pypi.

    Returns bool, message.

    The bool is True if we have the latest version, False if not.
    None if error was detected."""
    try:
        latest_version = get_latest_final_version()
    except Exception as e:

        return IsLatestVersion(
            None, f"Error checking snowfakery version: {type(e)}: {e}"
        )

    obsolete = latest_version > get_installed_version(version)
    if obsolete:
        return IsLatestVersion(
            False, f"An update to Snowfakery is available: {latest_version}"
        )
    else:
        return IsLatestVersion(True, "You have the latest version of Snowfakery")


def get_installed_version(version):
    """returns the version name (e.g. 2.0.0b58) that is installed"""
    return pkg_resources.parse_version(version)


# from cumulusci/utils/http/requests_utils.py
def safe_json_from_response(response):
    "Check JSON response is HTTP200 and actually JSON."
    response.raise_for_status()

    try:
        return response.json()
    except JSONDecodeError:
        raise ValueError(f"Cannot decode as JSON:  {response.text}")



import typing as T
import random
import math


class UpdatableRandomRange:
    def __init__(self, start: int, stop: int):
        assert stop > start
        self.min = start
        self._set_new_range_immediately(start, stop)

    def set_new_max(self, new_max: int):
        # do not replace RNG until old one is exhausted
        assert new_max >= self.cur_max
        self.cur_max = new_max

    def set_new_range(self, new_min: int, new_max: int):
        """Update the range subject to constraints

        There are two modes:

        If you update the range by changing only the top value,
        the generator will finish generating the first list before
        expanding its scope.

        So if you configured it with range(0,10) and then
        range(0,20) you would get

        shuffle(list(range(0,10)) + shuffle(list(range(10,20))

        Not:

        shuffle(list(range(0,10) + list(range(10,20))

        If you update the range by changing both values, the previous
        generator is just discarded, because you presumably don't
        want those values anymore. The new bottom must be higher
        than the old top. This preserves the rule that no value is
        ever produced twice.
        """
        if new_min == self.min:
            self.set_new_max(new_max)
        else:
            assert new_min >= self.orig_max, (new_min, self.orig_max)
            self._set_new_range_immediately(new_min, new_max)

    def _set_new_range_immediately(self, new_min: int, new_max: int):
        assert new_max > new_min
        self.min = new_min
        self.orig_max = self.cur_max = new_max
        self.num_generator = random_range(self.min, self.orig_max)

    def __iter__(self):
        return self

    def __next__(self):
        rv = next(self.num_generator, None)

        if rv is not None:
            return rv

        if self.cur_max <= self.orig_max:
            raise StopIteration()

        self.min = self.orig_max
        self.num_generator = random_range(self.min, self.cur_max)
        self.orig_max = self.cur_max
        return next(self.num_generator)


def random_range(start: int, stop: int) -> T.Generator[int, None, None]:
    """
    Return a randomized "range" using a Linear Congruential Generator
    to produce the number sequence. Parameters are the same as for
    python builtin "range".
        Memory  -- storage for 8 integers, regardless of parameters.
        Compute -- at most 2*"maximum" steps required to generate sequence.
    Based on https://stackoverflow.com/a/53551417/113477

    # Set a default values the same way "range" does.
    """
    step = 1  # step is hard-coded to "1" because it seemed to be buggy
    # and not important for our use-case

    # Use a mapping to convert a standard range into the desired range.
    def mapping(i):
        return (i * step) + start

    # Compute the number of numbers in this range.
    maximum = (stop - start) // step

    # Seed range with a random integer.
    value = random.randint(0, maximum)
    #
    # Construct an offset, multiplier, and modulus for a linear
    # congruential generator. These generators are cyclic and
    # non-repeating when they maintain the properties:
    #
    #   1) "modulus" and "offset" are relatively prime.
    #   2) ["multiplier" - 1] is divisible by all prime factors of "modulus".
    #   3) ["multiplier" - 1] is divisible by 4 if "modulus" is divisible by 4.
    #
    offset = random.randint(0, maximum) * 2 + 1  # Pick a random odd-valued offset.
    multiplier = (
        4 * (maximum // 4) + 1
    )  # Pick a multiplier 1 greater than a multiple of 4.
    modulus = int(
        2 ** math.ceil(math.log2(maximum))
    )  # Pick a modulus just big enough to generate all numbers (power of 2).
    # Track how many random numbers have been returned.
    found = 0
    while found < maximum:
        # If this is a valid value, yield it in generator fashion.
        if value < maximum:
            found += 1
            yield mapping(value)
        # Calculate the next value in the sequence.
        value = (value * multiplier + offset) % modulus



from base64 import b64encode
from snowfakery.plugins import SnowfakeryPlugin


class Base64(SnowfakeryPlugin):
    class Functions:
        def encode(self, data):
            return b64encode(bytes(str(data), "latin1")).decode("ascii")
